# 八、测试并发应用

在本章中，我们将介绍：

*   监控`Lock`接口
*   监控`Phaser`类
*   监视执行器框架
*   监视 Fork/Join 池
*   编写有效的日志消息
*   用 FindBugs 分析并发代码
*   配置 Eclipse 以调试并发代码
*   配置 NetBeans 以调试并发代码
*   用多线程技术测试并发代码

# 导言

测试应用是一项关键任务。在为最终用户准备好应用之前，您必须证明其正确性。您使用一个测试过程来证明实现了正确性，并且纠正了错误。测试阶段是任何软件开发中的常见任务，也是**质量保证**过程。您可以找到大量关于测试过程的文献，以及可以应用于开发的不同方法。还有很多库，比如`JUnit`和应用，比如 Apache`JMetter`，您可以使用它们以自动化的方式测试 Java 应用。在并发应用开发中，它甚至更为关键。

并发应用有两个或多个共享数据结构并相互交互的线程，这一事实给测试阶段增加了更多的难度。在测试并发应用时，您将面临的最大问题是线程的执行是不确定的。您无法保证线程的执行顺序，因此很难重现错误。

在本章中，您将学习：

*   如何获取有关并发应用中的元素的信息。这些信息可以帮助您测试并发应用。
*   如何使用 IDE（集成开发环境）和其他工具（如 FindBugs）来测试并发应用。
*   如何使用库（如 MultithreadedTC）自动化测试。

# 监控锁接口

`Lock`接口是 Java 并发 API 提供的获取代码块同步的基本机制之一。它允许定义**临界截面**。临界段是访问共享资源的代码块，不能由多个线程同时执行。该机制由`Lock`接口和`ReentrantLock`类实现。

在本食谱中，您将了解您可以获得关于`Lock`对象的哪些信息以及如何获得这些信息。

## 准备好了吗

此配方的示例已使用 EclipseIDE 实现。如果您使用 Eclipse 或其他 IDE（如 NetBeans），请打开它并创建一个新的 Java 项目。

## 怎么做。。。

按照以下步骤来实现该示例：

1.  创建一个名为`MyLock`的类来扩展`ReentrantLock`类。

    ```java
    public class MyLock extends ReentrantLock {
    ```

2.  执行`getOwnerName()`方法。此方法返回使用`Lock`类`getOwner()`的受保护方法控制锁（如果有）的线程的名称。

    ```java
      public String getOwnerName() {
        if (this.getOwner()==null) {
          return "None";
        }
        return this.getOwner().getName();
      }
    ```

3.  执行`getThreads()`方法。此方法返回使用`Lock`类`getQueuedThreads()`的受保护方法在锁中排队的线程列表。

    ```java
      public Collection<Thread> getThreads() {
        return this.getQueuedThreads();
      }
    ```

4.  创建一个名为`Task`的类，该类实现`Runnable`接口。

    ```java
    public class Task implements Runnable {
    ```

5.  声明名为`lock`的私有`Lock`属性。

    ```java
      private Lock lock;
    ```

6.  实现类的构造函数以初始化其属性。

    ```java
      public Task (Lock lock) {
        this.lock=lock;
      }
    ```

7.  执行`run()`方法。创建一个包含五个步骤的循环。

    ```java
      @Override
      public void run() {
        for (int i=0; i<5; i++) {
    ```

8.  使用`lock()`方法获取锁并打印消息。

    ```java
          lock.lock();
          System.out.printf("%s: Get the Lock.\n",Thread.currentThread().getName());
    ```

9.  将线程休眠 500 毫秒。使用`unlock()`方法释放锁并打印消息。

    ```java
          try {
            TimeUnit.MILLISECONDS.sleep(500);
            System.out.printf("%s: Free the Lock.\n",Thread.currentThread().getName());
          } catch (InterruptedException e) {
            e.printStackTrace();
          } finally {
            lock.unlock();
          }
            }
      }
    }
    ```

10.  通过使用`main()`方法创建名为`Main`的类来创建示例的主类。

    ```java
    public class Main {

      public static void main(String[] args) throws Exception {
    ```

11.  创建一个名为`lock`的`MyLock`对象。

    ```java
        MyLock lock=new MyLock();
    ```

12.  为五个`Thread`对象创建数组。

    ```java
        Thread threads[]=new Thread[5];
    ```

13.  创建并启动五个线程来执行五个`Task`对象。

    ```java
        for (int i=0; i<5; i++) {
          Task task=new Task(lock);
          threads[i]=new Thread(task);
          threads[i].start();
        }
    ```

14.  创建一个包含 15 个步骤的循环。

    ```java
        for (int i=0; i<15; i++) {
    ```

15.  在控制台中写入锁所有者的名称。

    ```java
          System.out.printf("Main: Logging the Lock\n");
          System.out.printf("************************\n");
          System.out.printf("Lock: Owner : %s\n",lock.getOwnerName());
    ```

16.  显示排队等待锁定的线程的编号和名称。

    ```java
    .out.printf("Lock: Queued Threads: %s\n",lock.hasQueuedThreads());
          if (lock.hasQueuedThreads()){
            System.out.printf("Lock: Queue Length: %d\n",lock.getQueueLength());
            System.out.printf("Lock: Queued Threads: ");
            Collection<Thread> lockedThreads=lock.getThreads();
            for (Thread lockedThread : lockedThreads) {
            System.out.printf("%s ",lockedThread.getName());
            }
            System.out.printf("\n");
          }
    ```

17.  显示`Lock`对象的公平性和状态信息。

    ```java
          System.out.printf("Lock: Fairness: %s\n",lock.isFair());
          System.out.printf("Lock: Locked: %s\n",lock.isLocked());
          System.out.printf("************************\n");
    ```

18.  让线程休眠 1 秒，然后关闭循环和类。

    ```java
          TimeUnit.SECONDS.sleep(1);
        }

      }

    }
    ```

## 它是如何工作的。。。

在此配方中，您实现了`MyLock`类，该类扩展了`ReentrantLock`类以返回否则不可用的信息—它是`ReentrantLock`类的受保护数据。`MyLock`类实现的方法有：

*   `getOwnerName()`：只有一个线程可以执行受`Lock`对象保护的临界段。锁存储执行临界段的线程。此线程由`ReentrantLock`类的受保护`getOwner()`方法返回。此方法使用`getOwner()`方法返回该线程的名称。
*   `getThreads()`：当线程正在执行临界段时，尝试进入该临界段的其他线程将处于休眠状态，直到它们可以继续执行该临界段。`ReentrantLock`类的受保护方法`getQueuedThreads()`返回等待执行临界段的线程列表。此方法返回由方法和`getQueuedThreads()`方法返回的结果。

我们还使用了在`ReentrantLock`类中实现的其他方法：

*   `hasQueuedThreads()`：此方法返回一个`Boolean`值，指示是否有线程等待获取此锁
*   `getQueueLength()`：此方法返回等待获取此锁的线程数
*   `isLocked()`：此方法返回一个`Boolean`值，指示此锁是否属于线程
*   `isFair()`：此方法返回一个`Boolean`值，指示此锁是否激活了公平模式

## 还有更多。。。

`ReentrantLock`类中还有其他方法可用于获取`Lock`对象的信息：

*   `getHoldCount()`：返回当前线程获取锁的次数
*   `isHeldByCurrentThread()`：返回一个`Boolean`值，指示锁是否属于当前线程

## 另见

*   [第 2 章](2.html "Chapter 2. Basic Thread Synchronization")*基本线程同步*中的*用锁同步代码块*配方
*   [第 7 章](7.html "Chapter 7. Customizing Concurrency Classes")*定制并发类*中*实现自定义锁类*配方

# 监控相量等级

Java 并发 API 提供的最复杂、功能最强大的功能之一是使用`Phaser`类执行并发阶段性任务。当我们有一些按步骤划分的并发任务时，这种机制很有用。`Phaser`类为我们提供了在每个步骤结束时同步线程的机制，因此在所有线程完成第一步之前，没有线程开始第二步。

在本食谱中，您将了解有关`Phaser`类状态的哪些信息以及如何获取这些信息。

## 准备好了吗

此配方的示例已使用 EclipseIDE 实现。如果您使用 Eclipse 或其他 IDE（如 NetBeans），请打开它并创建一个新的 Java 项目。

## 怎么做。。。

按照以下步骤来实现该示例：

1.  创建一个名为`Task`的类来实现`Runnable`接口。

    ```java
    public class Task implements Runnable {
    ```

2.  声明名为`time`的私有`int`属性。

    ```java
      private int time;
    ```

3.  声明名为`phaser`的私有`Phaser`属性。

    ```java
      private Phaser phaser;
    ```

4.  实现类的构造函数以初始化其属性。

    ```java
      public Task(int time, Phaser phaser) {
        this.time=time;
        this.phaser=phaser;
      }
    ```

5.  执行`run()`方法。首先，指示`phaser`属性任务使用`arrive()`方法开始执行。

    ```java
        @Override
      public void run() {

        phaser.arrive();
    ```

6.  在控制台中写入指示第一阶段开始的消息，按照`time`属性指定的秒数使线程休眠，在控制台中写入指示第一阶段结束的消息，并使用`phaser`属性的`arriveAndAwaitAdvance()`方法与其余任务同步。

    ```java
        System.out.printf("%s: Entering phase 1.\n",Thread.currentThread().getName());
        try {
          TimeUnit.SECONDS.sleep(time);
        } catch (InterruptedException e) {
          e.printStackTrace();
        }
        System.out.printf("%s: Finishing phase 1.\n",Thread.currentThread().getName());
        phaser.arriveAndAwaitAdvance();
    ```

7.  对第二和第三阶段重复该行为。在第三阶段结束时，使用`arriveAndDeregister()`方法代替`arriveAndAwaitAdvance()`。

    ```java
        System.out.printf("%s: Entering phase 2.\n",Thread.currentThread().getName());
        try {
          TimeUnit.SECONDS.sleep(time);
        } catch (InterruptedException e) {
          e.printStackTrace();
        }
        System.out.printf("%s: Finishing phase 2.\n",Thread.currentThread().getName());
        phaser.arriveAndAwaitAdvance();

        System.out.printf("%s: Entering phase 3.\n",Thread.currentThread().getName());
        try {
          TimeUnit.SECONDS.sleep(time);
        } catch (InterruptedException e) {
          e.printStackTrace();
        }
        System.out.printf("%s: Finishing phase 3.\n",Thread.currentThread().getName());

        phaser.arriveAndDeregister();
    ```

8.  通过使用`main()`方法创建名为`Main`的类，实现示例的主类。

    ```java
    public class Main {

      public static void main(String[] args) throws Exception {
    ```

9.  创建一个名为`phaser`的新`Phaser`对象，其中有三名参与者。

    ```java
        Phaser phaser=new Phaser(3);
    ```

10.  创建并启动三个线程以执行三个任务对象。

    ```java
        for (int i=0; i<3; i++) {
          Task task=new Task(i+1, phaser);
          Thread thread=new Thread(task);
          thread.start();
        }
    ```

11.  创建一个包含 10 个步骤的循环，以写入关于`phaser`对象的信息。

    ```java
        for (int i=0; i<10; i++) {
    ```

12.  填写有关注册方、相位器相位、到达方和未到达方的信息。

    ```java
        for (int i=0; i<10; i++) {
          System.out.printf("********************\n");
          System.out.printf("Main: Phaser Log\n");
          System.out.printf("Main: Phaser: Phase: %d\n",phaser.getPhase());
          System.out.printf("Main: Phaser: Registered Parties: %d\n",phaser.getRegisteredParties());
          System.out.printf("Main: Phaser: Arrived Parties: %d\n",phaser.getArrivedParties());
          System.out.printf("Main: Phaser: Unarrived Parties: %d\n",phaser.getUnarrivedParties());
          System.out.printf("********************\n");
    ```

13.  让线程休眠 1 秒，然后关闭循环和类。

    ```java
            TimeUnit.SECONDS.sleep(1);
        }
      }
    }
    ```

## 它是如何工作的。。。

在这个配方中，我们在`Task`类中实现了一个分阶段任务。此分阶段任务有三个阶段，使用`Phaser`接口与其他`Task`对象同步。main 类启动三个任务，当这些任务执行其阶段时，它将有关`phaser`对象状态的信息打印到控制台。我们使用以下方法获取`phaser`对象的状态：

*   `getPhase()`：此方法返回`phaser`对象的实际相位
*   `getRegisteredParties()`：此方法返回使用`phaser`对象作为同步机制的任务数
*   `getArrivedParties()`：此方法返回实际阶段结束时到达的任务数
*   `getUnarrivedParties()`：此方法返回实际阶段结束时尚未到达的任务数

以下屏幕截图显示了程序的部分输出：

![How it works...](img/7881_08_01.jpg)

## 另见

*   [第 3 章](3.html "Chapter 3. Thread Synchronization Utilities")*线程同步工具*中的*运行并发阶段任务*配方

# 监控执行器框架

Executor 框架提供了一种机制，将任务的执行与执行这些任务的线程创建和管理分离开来。如果使用执行器，则只需实现`Runnable`对象并将其发送给执行器即可。执行者负责管理线程。当您向执行者发送任务时，它会尝试使用池线程来执行此任务，以避免创建新线程。该机制由`Executor`接口及其实现类`ThreadPoolExecutor`类提供。

在本食谱中，您将了解您可以获得关于`ThreadPoolExecutor`执行人状态的哪些信息以及如何获得这些信息。

## 准备好了吗

此配方的示例已使用 EclipseIDE 实现。如果您使用 Eclipse 或其他 IDE（如 NetBeans），请打开它并创建一个新的 Java 项目。

## 怎么做。。。

按照以下步骤来实现该示例：

1.  创建一个名为`Task`的类，该类实现`Runnable`接口。

    ```java
    public class Task implements Runnable {
    ```

2.  声明名为`milliseconds`的私有`long`属性。

    ```java
      private long milliseconds;
    ```

3.  实现类的构造函数以初始化其属性。

    ```java
      public Task (long milliseconds) {
        this.milliseconds=milliseconds;
      }
    ```

4.  执行`run()`方法。按照`milliseconds`属性指定的毫秒数休眠线程。

    ```java
      @Override
      public void run() {

        System.out.printf("%s: Begin\n",Thread.currentThread().getName());
        try {
          TimeUnit.MILLISECONDS.sleep(milliseconds);
        } catch (InterruptedException e) {
          e.printStackTrace();
        }
        System.out.printf("%s: End\n",Thread.currentThread().getName());

      }
    ```

5.  通过使用`main()`方法创建名为`Main`的类来实现示例的主类。

    ```java
    public class Main {

      public static void main(String[] args) throws Exception {
    ```

6.  使用`Executors`类的`newCachedThreadPool()`方法创建新`Executor`对象。

    ```java
        ThreadPoolExecutor executor = (ThreadPoolExecutor)Executors.newCachedThreadPool();
    ```

7.  创建并向执行者提交 10 个`Task`对象。使用随机数初始化对象。

    ```java
        Random random=new Random();
        for (int i=0; i<10; i++) {
          Task task=new Task(random.nextInt(10000));
          executor.submit(task);
        }
    ```

8.  创建一个包含五个步骤的循环。在每个步骤中，写入有关调用`showLog()`方法的执行器的信息，并使线程休眠一秒钟。

    ```java
        for (int i=0; i<5; i++){
          showLog(executor);
          TimeUnit.SECONDS.sleep(1);
        }
    ```

9.  使用`shutdown()`方法关闭执行器。

    ```java
        executor.shutdown();
    ```

10.  在每个步骤中创建另一个包含五个步骤的循环，写入有关调用`showLog()`方法的执行器的信息，并使线程休眠一秒钟。

    ```java
        for (int i=0; i<5; i++){
          showLog(executor);
          TimeUnit.SECONDS.sleep(1);
        }
    ```

11.  使用`awaitTermination()`方法等待执行器的最终确定。

    ```java
          executor.awaitTermination(1, TimeUnit.DAYS);
    ```

12.  显示有关程序结束的消息。

    ```java
        System.out.printf("Main: End of the program.\n");
      }
    ```

13.  执行接收`Executor`作为参数的`showLog()`方法。写入有关池大小、任务数和执行器状态的信息。

    ```java
      private static void showLog(ThreadPoolExecutor executor) {
        System.out.printf("*********************");
        System.out.printf("Main: Executor Log");
        System.out.printf("Main: Executor: Core Pool Size: %d\n",executor.getCorePoolSize());
        System.out.printf("Main: Executor: Pool Size: %d\n",executor.getPoolSize());
        System.out.printf("Main: Executor: Active Count: %d\n",executor.getActiveCount());
        System.out.printf("Main: Executor: Task Count: %d\n",executor.getTaskCount());
        System.out.printf("Main: Executor: Completed Task Count: %d\n",executor.getCompletedTaskCount());
        System.out.printf("Main: Executor: Shutdown: %s\n",executor.isShutdown());
        System.out.printf("Main: Executor: Terminating: %s\n",executor.isTerminating());
        System.out.printf("Main: Executor: Terminated: %s\n",executor.isTerminated());
        System.out.printf("*********************\n");
      }
    ```

## 它是如何工作的。。。

在这个配方中，您已经实现了一个任务，该任务会将其执行线程阻塞随机数毫秒。然后，您已经向执行者发送了 10 个任务，在等待任务完成的同时，您已经将有关执行者状态的信息写入了控制台。您使用以下方法获取`Executor`对象的状态：

*   `getCorePoolSize()`：此方法返回一个`int`编号，即线程的核心编号。它是当执行器不执行任何任务时，内部线程池中的最小线程数。
*   `getPoolSize()`：此方法返回一个`int`值，即内部线程池的实际大小。
*   `getActiveCount()`：此方法返回一个`int`编号，即当前正在执行任务的线程数。
*   `getTaskCount()`：此方法返回一个`long`编号，即已计划执行的任务数。
*   `getCompletedTaskCount()`：此方法返回一个`long`编号，该编号是该执行者已执行并已完成执行的任务数。
*   `isShutdown()`：当执行器的`shutdown()`方法被调用以完成其执行时，此方法返回一个`Boolean`值。
*   `isTerminating()`：此方法在执行者进行`shutdown()`操作时返回一个`Boolean`值，但尚未完成。
*   `isTerminated()`：此方法在该执行器完成执行后返回一个`Boolean`值。

## 另见

*   [第 4 章](4.html "Chapter 4. Thread Executors")*线程执行器*中的*创建线程执行器*配方
*   [第 7 章](7.html "Chapter 7. Customizing Concurrency Classes")*定制并发类*中的*定制 ThreadPoolExecutor 类*配方
*   [第 7 章](7.html "Chapter 7. Customizing Concurrency Classes")*定制并发类*中*实现基于优先级的执行器类*配方

# 监控 Fork/Join 池

Executor 框架提供一种机制，允许将任务实现与执行这些任务的线程的创建和管理分离。Java 7 包括针对特定类型问题的 Executor 框架的扩展，该问题将提高其他解决方案的性能（如直接使用`Thread`对象或 Executor 框架）。这是 Fork/Join 框架。

该框架旨在通过使用`fork()`和`join()`操作的分治技术来解决那些可以分解为更小任务的问题。实现此行为的主类是`ForkJoinPool`类。

在本食谱中，您将了解您可以获得关于`ForkJoinPool`类的哪些信息以及如何获得这些信息。

## 准备好了吗

此配方的示例已使用 EclipseIDE 实现。如果您使用 Eclipse 或其他 IDE，如 NetBeans，请打开它并创建一个新的 Java 项目。

## 怎么做。。。

按照以下步骤来实现该示例：

1.  创建一个名为`Task`的类来扩展`RecursiveAction`类。

    ```java
    public class Task extends RecursiveAction{
    ```

2.  声明一个名为`array`的私有`int`数组属性来存储要递增的元素数组。

    ```java
    private int array[];
    ```

3.  声明两个名为`start`和`end`的私有`int`属性，以存储此任务必须处理的元素块的开始和结束位置。

    ```java
      private int start;
      private int end;
    ```

4.  实现类的构造函数以初始化其属性。

    ```java
      public Task (int array[], int start, int end) {
        this.array=array;
        this.start=start;
        this.end=end;
      }
    ```

5.  使用任务的主要逻辑执行`compute()`方法。如果任务必须处理 100 多个元素，则将该组元素分成两部分，创建两个任务来执行这些部分，然后使用`fork()`方法开始执行，然后使用`join()`方法等待其最终完成。

    ```java
      protected void compute() {
        if (end-start>100) {
          int mid=(start+end)/2;
          Task task1=new Task(array,start,mid);
          Task task2=new Task(array,mid,end);

          task1.fork();
          task2.fork();

          task1.join();
          task2.join();
    ```

6.  如果任务必须处理 100 个或更少的元素，则在每次操作后将线程休眠 5 毫秒，以增加这些元素。

    ```java
        } else {
          for (int i=start; i<end; i++) {
            array[i]++;

            try {
              Thread.sleep(5);
            } catch (InterruptedException e) {
              e.printStackTrace();
            }
          }
        }
      }
    }
    ```

7.  通过使用`main()`方法创建名为`Main`的类来实现示例的主类。

    ```java
    public class Main {

      public static void main(String[] args) throws Exception {
    ```

8.  创建一个名为`pool`的`ForkJoinPool`对象。

    ```java
        ForkJoinPool pool=new ForkJoinPool();
    ```

9.  创建一个名为`array`的整数数组，包含 10000 个元素。

    ```java
        int array[]=new int[10000];
    ```

10.  创建一个新的`Task`对象来处理整个数组。

    ```java
        Task task1=new Task(array,0,array.length);
    ```

11.  使用`execute()`方法在池中发送要执行的任务。

    ```java
        pool.execute(task1);
    ```

12.  当任务未完成执行时，调用`showLog()`方法写入`ForkJoinPool`类的状态信息，并让线程休眠一秒钟。

    ```java
        while (!task1.isDone()) {
          showLog(pool);
          TimeUnit.SECONDS.sleep(1);
        }
    ```

13.  使用`shutdown()`方法关闭池。

    ```java
        pool.shutdown();
    ```

14.  使用`awaitTermination()`方法等待池的最终确定。

    ```java
          pool.awaitTermination(1, TimeUnit.DAYS);
    ```

15.  调用`showLog()`方法写入`ForkJoinPool`类的状态信息，并在控制台中写入一条消息，表示程序结束。

    ```java
        showLog(pool);
        System.out.printf("Main: End of the program.\n");
    ```

16.  执行`showLog()`方法。它接收一个`ForkJoinPool`对象作为参数，并写入有关其状态以及正在执行的线程和任务的信息。

    ```java
      private static void showLog(ForkJoinPool pool) {
        System.out.printf("**********************\n");
        System.out.printf("Main: Fork/Join Pool log\n");
        System.out.printf("Main: Fork/Join Pool: Parallelism: %d\n",pool.getParallelism());
        System.out.printf("Main: Fork/Join Pool: Pool Size: %d\n",pool.getPoolSize());
        System.out.printf("Main: Fork/Join Pool: Active Thread Count: %d\n",pool.getActiveThreadCount());
        System.out.printf("Main: Fork/Join Pool: Running Thread Count: %d\n",pool.getRunningThreadCount());
        System.out.printf("Main: Fork/Join Pool: Queued Submission: %d\n",pool.getQueuedSubmissionCount());
        System.out.printf("Main: Fork/Join Pool: Queued Tasks: %d\n",pool.getQueuedTaskCount());
        System.out.printf("Main: Fork/Join Pool: Queued Submissions: %s\n",pool.hasQueuedSubmissions());
        System.out.printf("Main: Fork/Join Pool: Steal Count: %d\n",pool.getStealCount());
        System.out.printf("Main: Fork/Join Pool: Terminated : %s\n",pool.isTerminated());
        System.out.printf("**********************\n");
      }
    ```

## 它是如何工作的。。。

在这个配方中，您已经实现了一个任务，该任务使用`ForkJoinPool`类和扩展`RecursiveAction`类的`Task`类来增加数组的元素；您可以在`ForkJoinPool`类中执行的任务之一。当任务处理数组时，您可以将有关`ForkJoinPool`类状态的信息打印到控制台。您已使用以下方法获取`ForkJoinPool`类的状态：

*   `getPoolSize()`：此方法返回一个`int`值，即 fork-join 池内部池的工作线程数
*   `getParallelism()`：此方法返回为池建立的所需并行级别
*   `getActiveThreadCount()`：此方法返回当前正在执行任务的线程数
*   `getRunningThreadCount()`：此方法返回任何同步机制中未阻塞的工作线程数
*   `getQueuedSubmissionCount()`：此方法返回已提交到尚未开始执行的池的任务数
*   `getQueuedTaskCount()`：此方法返回已提交到已开始执行的池的任务数
*   `hasQueuedSubmissions()`：此方法返回一个`Boolean`值，指示此池是否已将尚未开始执行的任务排入队列
*   `getStealCount()`：此方法返回一个`long`值，该值表示工作线程从另一个线程窃取任务的次数
*   `isTerminated()`：此方法返回一个`Boolean`值，指示 fork/join 池是否已完成执行

## 另见

*   [第 5 章](5.html "Chapter 5. Fork/Join Framework")*Fork/Join 框架*中的*创建 Fork/Join 池*配方
*   *实现 ThreadFactory 接口为[第 7 章](7.html "Chapter 7. Customizing Concurrency Classes")*定制并发类*中的 Fork/Join 框架*配方生成自定义线程
*   [第 7 章](7.html "Chapter 7. Customizing Concurrency Classes")*定制并发类*中*定制运行在 Fork/Join 框架*配方中的任务

# 写入有效日志消息

**日志系统**是允许您将信息写入一个或多个目的地的机制。**记录器**具有以下组件：

*   **一个或多个处理程序**：一个处理程序将确定日志消息的目的地和格式。您可以将日志消息写入控制台、文件或数据库。
*   **名称**：通常，基于类名及其包名的类中使用的记录器的名称。
*   **级别**：日志消息有一个关联的级别，表示其重要性。记录器还有一个级别，用于决定要写入哪些消息。它只写与其级别同等重要或更重要的消息。

您应使用日志系统，主要用于以下两个目的：

*   捕获异常时，尽可能多地编写信息。这将有助于定位错误并解决它。
*   编写有关程序正在执行的类和方法的信息。

在本教程中，您将学习如何使用`java.util.logging`包提供的类向并发应用添加日志系统。

## 准备好了吗

此配方的示例已使用 EclipseIDE 实现。如果您使用 Eclipse 或其他 IDE（如 NetBeans），请打开它并创建一个新的 Java 项目。

## 怎么做。。。

按照以下步骤来实现该示例：

1.  创建一个名为`MyFormatter`的类来扩展`java.util.logging.Formatter`类。实现抽象的`format()`方法。它接收一个`LogRecord`对象作为参数，并返回一个带有日志消息的`String`对象。

    ```java
    public class MyFormatter extends Formatter {
      @Override
      public String format(LogRecord record) {

        StringBuilder sb=new StringBuilder();
        sb.append("["+record.getLevel()+"] - ");
        sb.append(new Date(record.getMillis())+" : "); 
          sb.append(record.getSourceClassName()+ "."+record.getSourceMethodName()+" : ");
        sb.append(record.getMessage()+"\n");.
        return sb.toString();
      }
    ```

2.  创建一个名为`MyLogger`的类。

    ```java
    public class MyLogger {
    ```

3.  声明一个名为`handler`的私有静态`Handler`属性。

    ```java
      private static Handler handler;
    ```

4.  实现 public static 方法`getLogger()`来创建用于写入日志消息的`Logger`对象。它接收一个名为`name`的`String`参数。

    ```java
      public static Logger getLogger(String name){
    ```

5.  使用`Logger`类的`getLogger()`方法获取与作为参数接收的名称关联的`java.util.logging.Logger`。

    ```java
        Logger logger=Logger.getLogger(name);
    ```

6.  建立日志级别，使用和`setLevel()`方法写入所有日志消息。

    ```java
        logger.setLevel(Level.ALL);
    ```

7.  如果 handler 属性具有`null`值，则创建一个新的`FileHandler`对象，将日志消息写入`recipe8.log`文件中。使用`setFormatter()`对象将`MyFormatter`对象指定给该处理程序作为格式化程序。

    ```java
        try {
          if (handler==null) {
            handler=new FileHandler("recipe8.log");
            Formatter format=new MyFormatter();
            handler.setFormatter(format);
          }
    ```

8.  如果`Logger`对象没有关联的处理程序，请使用`addHandler()`方法分配该处理程序。

    ```java
          if (logger.getHandlers().length==0) {
            logger.addHandler(handler);
          }
        } catch (SecurityException e) {
          e.printStackTrace();
        } catch (IOException e) {
          e.printStackTrace();
        }
    ```

9.  返回创建的`Logger`对象。

    ```java
        return logger;
      }
    ```

10.  创建一个名为`Task`的类来实现`Runnable`接口。这将是用来测试你的`Logger`对象的任务。

    ```java
    public class Task implements Runnable {
    ```

11.  执行`run()`方法。

    ```java
      @Override
      public void run() {
    ```

12.  首先，声明一个名为`logger`的`Logger`对象。使用`MyLogger`类的`getLogger()`方法将该类的名称作为参数进行初始化。

    ```java
        Logger logger= MyLogger.getLogger(this.getClass().getName());
    ```

13.  使用`entering()`方法写入日志消息，指示开始执行该方法。

    ```java
        logger.entering(Thread.currentThread().getName(), "run()");
    Sleep the thread for two seconds.
        try {
          TimeUnit.SECONDS.sleep(2);
        } catch (InterruptedException e) {
          e.printStackTrace();
        }
    ```

14.  使用`exiting()`方法写入一条日志消息，指示该方法的执行结束。

    ```java
        logger.exiting(Thread.currentThread().getName(), "run()",Thread.currentThread());
      }
    ```

15.  通过使用`main()`方法创建名为`Main`的类来实现示例的主类。

    ```java
    public class Main {

      public static void main(String[] args) {
    ```

16.  声明一个名为`logger`的`Logger`对象。使用`MyLogger`类的`getLogger()`方法将字符串`Core`作为参数进行初始化。

    ```java
        Logger logger=MyLogger.getLogger("Core");
    ```

17.  写入日志消息，指示使用`entering()`方法开始执行主程序。

    ```java
        logger.entering("Core", "main()",args);
    ```

18.  创建一个`Thread`数组来存储五个线程。

    ```java
        Thread threads[]=new Thread[5];
    ```

19.  创建五个`Task`对象和五个线程来执行它们。编写日志消息以指示您将要启动一个新线程，并指示您已经创建了该线程。

    ```java
        for (int i=0; i<threads.length; i++) {
          logger.log(Level.INFO,"Launching thread: "+i);
          Task task=new Task();
          threads[i]=new Thread(task);
          logger.log(Level.INFO,"Thread created: "+ threads[i].getName());
          threads[i].start();
        }
    ```

20.  写一条日志消息，表明您已经创建了线程。

    ```java
        logger.log(Level.INFO,"Ten Threads created."+
    "Waiting for its finalization");
    ```

21.  等待使用`join()`方法完成五个线程。完成每个线程后，编写一条日志消息，指示线程已完成。

    ```java
        for (int i=0; i<threads.length; i++) {
          try {
            threads[i].join();
            logger.log(Level.INFO,"Thread has finished its execution",threads[i]);
          } catch (InterruptedException e) {
            logger.log(Level.SEVERE, "Exception", e);
          }
        }
    ```

22.  使用`exiting()`方法写入日志消息，以指示主程序的执行结束。

    ```java
        logger.exiting("Core", "main()");
      }
    ```

## 它是如何工作的。。。

在此配方中，您使用了为 Java 日志 API 提供的`Logger`类在并发应用中编写日志消息。首先，您已经实现了`MyFormatter`类，为日志消息提供了一种格式。此类扩展了声明抽象方法`format()`的`Formatter`类。此方法接收包含日志消息所有信息的`LogRecord`对象，并返回格式化的日志消息。在您的类中，您使用了`LogRecord`类的以下方法来获取有关日志消息的信息：

*   `getLevel()`：返回消息的级别
*   `getMillis()`：返回向`Logger`对象发送消息的日期
*   `getSourceClassName()`：返回向记录器发送消息的类的名称
*   `getSourceMessageName()`：返回向记录器发送消息的方法的名称

`getMessage()`返回日志消息。`MyLogger`类实现静态方法`getLogger()`，该方法创建`Logger`对象并分配`Handler`对象，以使用`MyFormatter`格式化程序将应用的日志消息写入`recipe8.log`文件。您使用该类的静态方法`getLogger()`创建`Logger`对象。此方法为作为参数传递的每个名称返回不同的对象。您只创建了一个`Handler`对象，因此所有`Logger`对象将在同一文件中写入其日志消息。您还将记录器配置为写入所有日志消息，无论其级别如何。

最后，您实现了一个`Task`对象和一个在日志文件中写入不同日志消息的主程序。您使用了以下方法：

*   `entering()`：写一条`FINER`级别的消息，表示方法开始执行
*   `exiting()`：写入一条`FINER`级别的消息，表示方法结束执行
*   `log()`：写一条指定级别的报文

## 还有更多。。。

当您使用日志系统时，您必须考虑两个要点：

*   **写下必要的信息**：如果你写的信息太少，记录器将不会有用，因为它无法实现其目的。如果您写入的信息太多，将生成太大的日志文件，这些文件将无法管理，并且很难获取必要的信息。
*   **对消息**使用适当的级别：如果您使用较高级别编写信息消息或使用较低级别编写错误消息，您将混淆查看日志文件的用户。要知道在错误情况下发生了什么会更加困难，或者您将有太多的信息无法知道错误的主要原因。

还有其他提供比`java.util.logging`包更完整的日志系统的库，例如 Log4j 或 slf4j 库。但是`java.util.logging`包是 JavaAPI 的一部分，它的所有方法都是多线程安全的，因此我们可以在并发应用中使用它而不会出现问题。

## 另见

*   [第 6 章](6.html "Chapter 6. Concurrent Collections")*并发集合*中的*使用非阻塞线程安全列表*配方
*   [第 6 章](6.html "Chapter 6. Concurrent Collections")*并发集合*中的*使用阻塞线程安全列表*配方
*   [第 6 章](6.html "Chapter 6. Concurrent Collections")、*并发集合*中的*使用按优先级*排序的阻塞线程安全列表
*   [第 6 章](6.html "Chapter 6. Concurrent Collections")*并发集合*中的*使用带延迟元素的线程安全列表*配方
*   [第 6 章](6.html "Chapter 6. Concurrent Collections")*并发集合*中的*使用线程安全导航地图*配方
*   [第 6 章](6.html "Chapter 6. Concurrent Collections")*并发集合*中的*生成并发随机数*配方

# 使用 FindBugs 分析并发代码

**静态代码分析工具**是一组分析应用源代码以查找潜在错误的工具。这些工具，如 Checkstyle、PMD 或 FindBugs 具有一组预定义的良好实践规则，并解析源代码以查找违反这些规则的情况。目标是在生产中执行之前，尽早发现导致性能低下的错误或位置。编程语言通常提供这样的工具，Java 也不例外。分析 Java 代码的工具之一是 FindBugs。它是一个开源工具，包含一系列分析 Java 并发代码的规则。

在本教程中，您将学习如何使用此工具分析 Java 并发应用。

## 准备好了吗

在说明此配方之前，您应该从项目的网页（[下载 FindBugshttp://findbugs.sourceforge.net/](http://findbugs.sourceforge.net/) ）。您可以下载独立应用或 Eclipse 插件。在此配方中，您将使用独立版本。

## 怎么做。。。

按照以下步骤来实现该示例：

1.  创建一个名为`Task`的类来扩展`Runnable`接口。

    ```java
    public class Task implements Runnable {
    ```

2.  声明名为`Lock`的私有`ReentrantLock`属性。

    ```java
      private ReentrantLock lock;
    ```

3.  实现类的构造函数。

    ```java
      public Task(ReentrantLock lock) {
        this.lock=lock;
      }
    ```

4.  执行`run()`方法。控制锁，使线程休眠 2 秒钟，然后释放锁。

    ```java
      @Override
      public void run() {
        lock.lock();
        try {
          TimeUnit.SECONDS.sleep(1);
          lock.unlock();
        } catch (InterruptedException e) {
          e.printStackTrace();
        }
      }
    ```

5.  通过使用`main()`方法创建名为`Main`的类来创建示例的主类。

    ```java
    public class Main {

      public static void main(String[] args) {
    ```

6.  声明并创建一个名为`lock`的`ReentrantLock`对象。

    ```java
        ReentrantLock lock=new ReentrantLock();
    ```

7.  创建 10 个`Task`对象和 10 个线程来执行这些任务。启动调用`run()`方法的线程。

    ```java
        for (int i=0; i<10; i++) {
          Task task=new Task(lock);
          Thread thread=new Thread(task);
          thread.run();
        }
      }
    ```

8.  将项目导出为`jar`文件。叫它`recipe8.jar`。使用 IDE 的菜单选项或`javac`和`jar`命令编译和压缩应用。
9.  启动 FindBugs 独立应用，在 Windows 中运行`findbugs.bat`**命令，或在 Linux 中运行`findbugs.sh`命令。**
***   Create a new project with the **New Project** option of the **File** menu in the menu bar.

    ![How to do it...](img/7881_08_02.jpg)

    *   The **FindBugs** application shows a window to configure the project. In the **Project Name** field introduce the text `Recipe08`. In the **Classpath for analysis** field add the `jar` file with the project and in the **Source directories** field add the directory with the source code of the example. Refer to the following screenshot:

    ![How to do it...](img/7881_08_03.jpg)

    *   点击**分析**按钮创建新项目并分析其代码。*   **FindBugs**应用显示了代码分析的结果。在本例中，它发现了两个 bug。*   点击其中一个 bug，您将在右侧面板中看到该 bug 的源代码，并在屏幕底部的面板中看到该 bug 的描述。**

 **## 它是如何工作的。。。

以下屏幕截图显示 FindBugs 的分析结果：

![How it works...](img/7881_08_04.jpg)

分析检测到应用中存在以下两个潜在错误：

*   类`Task`的`run()`方法中的一个。如果抛出了一个`InterruptedExeption`异常，任务不会释放锁，因为它不会执行`unlock()`方法。这可能会导致应用出现死锁情况。
*   另一个是在`Main`类的`main()`方法中，因为您直接调用了线程的`run()`方法，而不是`start()`方法来开始执行线程。

如果您双击两个 bug 中的一个，您将看到关于它的详细信息。由于您已经在项目的配置中包含了源代码参考，您还将看到检测到错误的源代码。下面的屏幕截图显示了一个示例：

![How it works...](img/7881_08_05.jpg)

## 还有更多。。。

请注意，FindBugs 只能检测一些有问题的情况（与并发代码相关或不相关）。例如，如果您删除`Task`类的`run()`方法中的`unlock()`调用并重复分析，FindBugs 不会提醒您在任务中获得锁，但您永远不会释放它。

使用静态代码分析工具帮助提高代码质量，但不要期望检测到代码中的所有错误。

## 另见

*   [第 8 章](8.html "Chapter 8. Testing Concurrent Applications")*测试并发应用*中的*配置 NetBeans 调试并发代码*配方

# 配置 Eclipse 调试并发代码

如今，几乎每一位程序员，不管使用哪种编程语言，都会用 IDE 创建他们的应用。它们提供了许多集成在同一应用中的有趣功能，例如：

*   项目管理
*   自动代码生成
*   自动文档生成
*   与控制版本系统的集成
*   用于测试应用的调试器
*   创建应用项目和元素的不同向导

IDE 最有用的特性之一是调试器。您可以一步一步地执行应用，并分析程序中所有对象和变量的值。

如果您使用 Java 编程语言，Eclipse 是最流行的 IDE 之一。它有一个集成的调试器，允许您测试应用。默认情况下，当调试并发应用并且调试器找到断点时，它只停止具有该断点的线程，而其余线程继续执行。

在此配方中，您将学习如何更改该配置以帮助您测试并发应用。

## 准备好了吗

您必须已经安装了 EclipseIDE。打开它并选择一个项目，其中实现了一个并发应用，例如，书中实现的一个配方。

## 怎么做。。。

按照以下步骤来实现该示例：

1.  选择菜单选项**窗口****首选项**。
2.  在左侧菜单中，展开**Java**选项。
3.  In the left-hand side menu, select the **Debug** option. The following screenshot shows the appearance of that window:

    ![How to do it...](img/7881_08_06.jpg)

4.  将新断点的**默认挂起策略**的值从**挂起线程**更改为**挂起 VM**（屏幕截图中用红色标记）。
5.  点击**确定**按钮确认变更。

## 它是如何工作的。。。

正如我们在本配方介绍中提到的，默认情况下，当您在 Eclipse 中调试并发 Java 应用时，调试进程发现断点，它只挂起首先命中断点的线程，而其他线程继续执行。以下屏幕截图显示了该情况的示例：

![How it works...](img/7881_08_07.jpg)

您可以看到，只有**worker-21**被挂起（在屏幕截图中标记为红色），而其余的线程正在运行。但是，如果将新断点的**默认挂起策略**更改为**挂起 VM**，则在调试并发应用时，所有线程都会挂起它们的执行，并且调试过程会遇到断点。以下屏幕截图显示了这种情况的示例：

![How it works...](img/7881_08_08.jpg)

通过更改，您可以看到所有线程都被挂起。您可以继续调试任意线程。选择最适合您需要的挂起策略。

# 配置 NetBeans 调试并发代码

在当今世界，软件对于开发能够正常工作、符合公司质量标准、且在未来可以在有限的时间内以尽可能低的成本轻松修改的应用是必不可少的。为了实现这一目标，必须使用一个 IDE，该 IDE 在一个公共接口下集成多个促进应用开发的工具（编译器和调试器）。

如果您使用 Java 编程语言，NetBeans 是最流行的 IDE 之一。它有一个集成的调试器，允许您测试您的应用。

在此配方中，您将学习如何更改该配置以帮助您测试并发应用。

## 准备好了吗

您应该安装 NetBeans IDE。打开它并创建一个新的 Java 项目。

## 怎么做。。。

按照以下步骤来实现该示例：

1.  创建一个名为`Task1`的类，并指定它实现`Runnable`接口。

    ```java
    public class Task1 implements Runnable {
    ```

2.  声明两个名为`lock1`和`lock2`的私有`Lock`属性。

    ```java
        private Lock lock1, lock2;
    ```

3.  实现类的构造函数以初始化其属性。

    ```java
        public Task1 (Lock lock1, Lock lock2) {
            this.lock1=lock1;
            this.lock2=lock2;
        }
    ```

4.  执行`run()`方法。首先，使用`lock()`方法获得`lock1`对象的控制，并在控制台中写入一条消息，表明您已经获得了它。

    ```java
     @Override
        public void run() {
            lock1.lock();
            System.out.printf("Task 1: Lock 1 locked\n");
    ```

5.  然后，使用`lock()`方法获得`lock2`对象的控制，并在控制台中写入一条消息，表明您已经获得了它。

    ```java
            lock2.lock();
            System.out.printf("Task 1: Lock 2 locked\n");
    Finally, release the two lock objects. First, the lock2 object and then the lock1 object.
            lock2.unlock();
            lock1.unlock();
        }
    ```

6.  创建一个名为`Task2`的类，并指定它实现`Runnable`接口。

    ```java
    public class Task2 implements Runnable{
    ```

7.  声明两个名为`lock1`和`lock2`的私有`Lock`属性。

    ```java
        private Lock lock1, lock2;
    ```

8.  实现类的构造函数来初始化其属性。

    ```java
        public Task2(Lock lock1, Lock lock2) {
            this.lock1=lock1;
            this.lock2=lock2;
        }
    ```

9.  执行`run()`方法。首先，使用`lock()`方法获得`lock2`对象的控制权，并在控制台中写入一条消息，表明您已经获得了它。

    ```java
     @Override
        public void run() {
            lock2.lock();
            System.out.printf("Task 2: Lock 2 locked\n");
    ```

10.  然后，使用`lock()`方法获得`lock1`对象的控制，并在控制台中写入一条消息，表明您已经获得了它。

    ```java
            lock1.lock();
            System.out.printf("Task 2: Lock 1 locked\n");
    ```

11.  最后，释放两个锁定对象。首先是`lock1`对象，然后是`lock2`对象。

    ```java
            lock1.unlock();
            lock2.unlock();
        }
    ```

12.  通过创建一个名为`Main`的类并向其添加`main()`方法来实现示例的主类。

    ```java
    public class Main {
    ```

13.  创建两个名为`lock1`和`lock2`的锁对象。

    ```java
            Lock lock1, lock2;
            lock1=new ReentrantLock();
            lock2=new ReentrantLock();
    ```

14.  创建一个名为`task1`的`Task1`对象。

    ```java
            Task1 task1=new Task1(lock1, lock2);
    ```

15.  创建一个名为`task2`的`Task2`对象。

    ```java
            Task2 task2=new Task2(lock1, lock2);
    ```

16.  使用两个线程执行这两个任务。

    ```java
            Thread thread1=new Thread(task1);
            Thread thread2=new Thread(task2);

            thread1.start();
            thread2.start();
    ```

17.  当这两个任务尚未完成执行时，每 500 毫秒在控制台中写入一条消息。使用`isAlive()`方法检查线程是否已完成执行。

    ```java
            while ((thread1.isAlive()) &&(thread2.isAlive())) {
                System.out.println("Main: The example is"+ "running");
                try {
                    TimeUnit.MILLISECONDS.sleep(500);
                } catch (InterruptedException ex) {
                    ex.printStackTrace();
                }
            }
    ```

18.  在`Task1`类的`run()`方法的`println()`方法的第一次调用中添加断点。
19.  Debug the program. You will see the **Debugging** window in the top left-hand side corner of the main NetBeans window. The next screenshot presents the appearance of that window with the thread that executes the `Task1` object slept because they have arrived at the breakpoint and the other threads running:

    ![How to do it...](img/7881_08_09.jpg)

20.  Pause the execution of the main thread. Select that thread, right-click, and select the **Suspend** option. The following screenshot shows the new appearance of the **Debugging** window. Refer to the following screenshot:

    ![How to do it...](img/7881_08_10.jpg)

21.  恢复两个暂停的线程。选择每个线程，右键单击，然后选择**恢复**选项。

## 它是如何工作的。。。

使用 NetBeans 调试并发应用时，当调试器遇到断点时，它会挂起遇到断点的线程，在左上角显示**调试**窗口，其中显示了当前正在运行的线程。

您可以使用该窗口使用**暂停**或**恢复**选项暂停或恢复当前正在运行的线程。您还可以使用**变量**选项卡查看线程的变量值或属性。

NetBeans 还包括死锁检测器。当您在**调试**菜单中选择**检查死锁**选项时，NetBeans 会对您正在调试的应用进行分析，以确定是否存在死锁情况。此示例显示了一个明显的死锁。第一个线程先得到锁`lock1`，然后得到锁`lock2`。第二个线程以相反的方式获得锁。插入的断点会引发死锁，但如果使用 NetBeans 死锁检测器，将找不到任何东西，因此应谨慎使用此选项。通过`synchronized`关键字更改两个任务中使用的锁，然后再次调试程序。`Task1`的代码如下所示：

```java
    @Override
    public void run() {
        synchronized(lock1) {
            System.out.printf("Task 1: Lock 1 locked\n");
            synchronized(lock2) {
                System.out.printf("Task 1: Lock 2 locked\n");
            }
        }
    } 
```

`Task2`类的代码与此类似，但会更改锁的顺序。如果再次调试示例，将再次获得死锁，但在这种情况下，死锁检测器会检测到死锁，如下图所示：

![How it works...](img/7881_08_11.jpg)

## 还有更多。。。

有控制调试器的选项。在**工具**菜单中选择**选项**选项。然后，选择**杂项**选项和**Java 调试器**选项卡。以下屏幕截图显示了该窗口的外观：

![There's more...](img/7881_08_12.jpg)

在窗口上有两个选项控制前面描述的行为：

*   **新断点挂起**：使用此选项，您可以配置 NetBeans 的行为，它在线程中查找断点。只能挂起具有断点的线程或应用的所有线程。
*   **步骤恢复**：使用此选项，您可以在恢复线程时配置 NetBeans 的行为。您只能恢复当前线程或所有线程。

这两个选项都已在前面显示的屏幕截图中标记。

## 另见

*   [第 8 章](8.html "Chapter 8. Testing Concurrent Applications")*测试并发应用*中的*配置 Eclipse 调试并发代码*配方

# 用多线程测试并发代码

MultithreadedTC 是一个用于测试并发应用的 Java 库。它的主要目标是解决并发应用不确定性的问题。你无法控制他们的执行顺序。为此，它包括一个内部**节拍器**，用于控制构成应用的不同线程的执行顺序。这些测试线程被实现为类的方法。

在本配方中，您将学习如何使用多线程 EDTC 库实现对`LinkedTransferQueue`的测试。

## 准备好了吗

您还必须从[下载多线程数据库 http://code.google.com/p/multithreadedtc/](http://code.google.com/p/multithreadedtc/) 和 JUnit 库，来自[的版本 4.10http://www.junit.org/](http://www.junit.org/) 。将文件`junit-4.10.jar`和`MultithreadedTC-1.01.jar`添加到项目的库中。

## 怎么做。。。

按照以下步骤来实现该示例：

1.  创建一个名为`ProducerConsumerTest`的类来扩展`MultithreadedTestCase`类。

    ```java
    public class ProducerConsumerTest extends MultithreadedTestCase {
    ```

2.  声明一个私有的`LinkedTransferQueue`属性，该属性由名为`queue`的`String`类参数化。

    ```java
      private LinkedTransferQueue<String> queue;
    ```

3.  执行`initialize()`方法。此方法不会接收任何参数，也不会返回任何值。它调用其父类的`initialize()`方法，然后初始化队列属性。

    ```java
       @Override
      public void initialize() {
        super.initialize();
        queue=new LinkedTransferQueue<String>();
        System.out.printf("Test: The test has been initialized\n");
      }
    ```

4.  执行`thread1()`方法。它将实现第一个使用者的逻辑。调用队列的`take()`方法，然后将返回值写入控制台。

    ```java
      public void thread1() throws InterruptedException {
        String ret=queue.take();
        System.out.printf("Thread 1: %s\n",ret);
      }
    ```

5.  `thread2()`实现方法。它将实现第二个消费者的逻辑。首先，使用`waitForTick()`方法等待第一个线程在`take()`方法中休眠。然后调用队列的`take()`方法，将返回值写入控制台。

    ```java
      public void thread2() throws InterruptedException {
        waitForTick(1);
        String ret=queue.take();
        System.out.printf("Thread 2: %s\n",ret);
      }
    ```

6.  `thread3()`实现方法。它将实现生产者的逻辑。首先，使用`waitForTick()`方法两次等待`take()`方法阻塞两个消费者。然后调用队列的`put()`方法，在队列中插入两个`Strings`。

    ```java
      public void thread3() {
        waitForTick(1);
        waitForTick(2);
        queue.put("Event 1");
        queue.put("Event 2");
        System.out.printf("Thread 3: Inserted two elements\n");
      }
    ```

7.  最后，实施`finish()`方法。在控制台中编写一条消息，指示测试已完成执行。使用`assertEquals()`方法检查两个事件是否已被消耗（因此队列大小为`0`。

    ```java
      public void finish() {
        super.finish();
        System.out.printf("Test: End\n");
        assertEquals(true, queue.size()==0);
        System.out.printf("Test: Result: The queue is empty\n");
      }
    ```

8.  通过使用`main()`方法创建名为`Main`的类来实现示例的主类。

    ```java
    public class Main {

      public static void main(String[] args) throws Throwable {
    ```

9.  创建一个名为`test`的`ProducerConsumerTest`对象。

    ```java
        ProducerConsumerTest test=new ProducerConsumerTest();
    ```

10.  使用`TestFramework`类的`runOnce()`方法执行测试。

    ```java
         System.out.printf("Main: Starting the test\n");
        TestFramework.runOnce(test);
        System.out.printf("Main: The test has finished\n");
    ```

## 它是如何工作的。。。

在此配方中，您已经使用多线程 DC 库为`LinkedTransferQueue`类实现了一个测试。您可以使用此库及其 metronome 对任何并发应用或类实施测试。在本例中，您已经用两个消费者和一个生产者实现了经典的生产者/消费者问题。您想测试在缓冲区中引入的第一个`String`对象是否被到达缓冲区的第一个消费者消费，在缓冲区中引入的第二个`String`对象是否被到达缓冲区的第二个消费者消费。

多线程 DC 库基于 JUnit 库，JUnit 库是 Java 中实现单元测试最常用的库。要使用 MultithreadedTC 库实现基本测试，必须扩展`MultithreadedTestCase`类。该类扩展了`junit.framework.AssertJUnit`类，该类包括检查测试结果的所有方法。它没有扩展`junit.framework.TestCase`类，因此不能将多线程的测试与其他 JUnit 测试集成。

然后，您可以实现以下方法：

*   `initialize()`：此方法的实现是可选的。它在启动测试时执行，因此您可以使用它初始化正在使用测试的对象。
*   `finish()`：此方法的实现是可选的。它在测试完成后执行。您可以使用它来关闭或释放测试期间使用的资源，或者检查测试结果。
*   实现测试的方法：这些方法具有您实现的测试的主要逻辑。它们必须以`thread`关键字开头，后跟字符串。例如，`thread1()`。

要控制线程的执行顺序，可以使用`waitForTick()`方法。此方法接收一个`integer`类型作为参数，并将执行该方法的线程置于睡眠状态，直到阻塞测试中运行的所有线程。当它们被阻塞时，多线程 DC 库将恢复通过调用`waitForTick()`方法阻塞的线程。

作为`waitForTick()`方法的参数传递的整数用于控制执行顺序。多线程的 metronome 库有一个内部计数器。当所有线程都被阻塞时，库会将计数器增加到被阻塞的`waitForTick()`调用中指定的下一个数字。

在内部，当多线程数据库必须执行测试时，它首先执行`initialize()`方法。然后，它为每个方法创建一个以`thread`关键字开头的线程（在您的示例中，是方法`thread1()`、`thread2()`和`thread3()`），当所有线程都完成执行后，它执行`finish()`方法。为了执行测试，您使用了`TestFramework`类的`runOnce()`方法。

## 还有更多。。。

如果 MultithreadedTC 库检测到测试的所有线程都被阻塞，但在`waitForTick()`方法中没有一个线程被阻塞，那么测试将被声明为处于死锁状态，并引发`java.lang.IllegalStateException`异常。

## 另见

*   [第 8 章](8.html "Chapter 8. Testing Concurrent Applications")*测试并发应用*中的*使用 FindBugs*配方分析并发代码**