# 七、测试 RxJava 应用

在编写软件时，尤其是编写将被许多用户使用的软件时，我们需要确保一切都正常工作。我们可以编写可读、结构良好、模块化的代码，这将使更改和维护更容易。我们应该编写测试，因为每个特性都有回归的危险。当我们已经对现有代码进行了测试时，重构就不会那么困难了，因为测试可以针对新的、更改过的代码运行。

几乎所有的东西都需要测试和自动化。甚至还有**测试驱动开发**（**TDD**）和**行为驱动开发**（**BDD**等意识形态。如果我们不编写自动化测试，我们不断变化的代码往往会随着时间的推移而中断，测试和维护变得更加困难。

在本章中，我们将不讨论为什么需要测试代码。我们会接受这是强制性的，是我们作为程序员生活的一部分。我们将学习如何测试使用 RxJava 编写的代码。

我们将看到，为它编写单元测试并不是那么困难，但是有一些难以测试的案例，例如*异步*`Observable`实例。我们将学习一些新的操作符，这将帮助我们进行测试和一种新的`Observable`实例。

话虽如此，我们将在本章中介绍以下内容：

*   通过`BlockingObservable`类和*聚合*操作符测试`Observable`实例
*   使用`TestSubscriber`实例进行深入测试
*   `TestScheduler`类和测试*异步*`Observable`实例

# 使用简单订阅进行测试

我们可以通过订阅*源*`Observable`实例并收集所有传入通知来测试我们得到了什么。为了证明这一点，我们将开发一个`factory`方法来创建一个新的`Observable`实例，并测试其行为。

该方法将接收一个`Comparator`实例和多个项，并返回`Observable`实例，将这些项按排序顺序发出。项目将根据传递的`Comparator`实例进行排序。

我们可以使用 TDD 开发该方法。让我们首先定义测试，如下所示：

```java
public class SortedObservableTest {
  private Observable<String> tested;
  private List<String> expected;
  @Before
  public void before() {
    tested = CreateObservable.<String>sorted(
 (a, b) -> a.compareTo(b),
 "Star", "Bar", "Car", "War", "Far", "Jar");
    expected = Arrays.asList(
      "Bar", "Car", "Far", "Jar", "Star", "War"
    );
  }
  TestData data = new TestData();
  tested.subscribe(
    (v) -> data.getResult().add(v),
    (e) -> data.setError(e),
    () -> data.setCompleted(true)
  );
  Assert.assertTrue(data.isCompleted());
  Assert.assertNull(data.getError());
  Assert.assertEquals(expected, data.getResult());
}
```

### 注

本章的示例使用**JUnit**框架进行测试。您可以在[找到更多关于此的信息 http://junit.org](http://junit.org) 。

测试使用两个变量来存储预定义的可重用状态。第一个是我们用作测试源的`Observable`实例。在设置`@Before`方法中，分配给我们方法`CreateObservable.sorted(Comparator, T...)`的结果，该结果尚未实现。我们比较一组`String`实例，并希望它们按照存储在*预期*变量第二个可重用字段中的顺序接收。

测试本身非常冗长。它使用`TestData`类的一个实例来存储从*测试的*`Observable`实例传入的通知。

如果有`OnCompleted`通知，`data.completed`字段设置为`True`。我们期望这种情况发生，这就是为什么我们在测试方法结束时断言它。如果有`OnError`通知，`data.error`字段设置为错误。我们不期望这种情况发生，所以我们断言它是`null`。

`Observable`实例发出的每个传入项都被添加到`data.resultList`字段。最后，它应该等于*预期的*`List`变量，我们断言。

### 注

上述测试的源代码可在[查看/下载 https://github.com/meddle0x53/learning-rxjava/blob/master/src/test/java/com/packtpub/reactive/chapter07/SortedObservableTest.java](https://github.com/meddle0x53/learning-rxjava/blob/master/src/test/java/com/packtpub/reactive/chapter07/SortedObservableTest.java) -这是第一种测试方法。

但是，这个测试当然失败了，因为`CreateObservable.sorted(Comparator, T...)`方法还没有实现。让我们实现它并再次运行测试：

```java
@SafeVarargs
public static <T> Observable<T> sorted(
  Comparator<? super T> comparator,
  T... data) {
    List<T> listData = Arrays.asList(data);
    listData.sort(comparator);
  return Observable.from(listData);
}
```

就这么简单！它只是将传递的`varargs`数组转换为`List`变量，并使用其`sort()`方法将其与传递的`Comparator`实例进行排序。然后，使用`Observable.from(Iterable)`方法返回所需的`Observable`实例。

### 注

上述实现的源代码位于：[https://github.com/meddle0x53/learning-rxjava/blob/master/src/main/java/com/packtpub/reactive/common/CreateObservable.java#L262](https://github.com/meddle0x53/learning-rxjava/blob/master/src/main/java/com/packtpub/reactive/common/CreateObservable.java#L262) 。

如果我们现在运行测试，它将通过。这很好！我们进行了第一次测试！但是编写类似的测试需要大量的样板代码。我们总是需要这三个状态变量，我们总是需要断言相同的东西。那么*异步*`Observable`实例呢，比如`interval()`和`timer()`方法创建的实例？

有一些删除样板变量的技术，稍后，我们将研究如何测试*异步*行为。现在，我们将介绍一种新类型的可观测。

# 阻塞可观测类

每个`Observable`实例都可以通过`toBlocking()`方法转换成`BlockingObservable`实例。`BlockingObservable`实例有多个方法阻塞当前线程，而*源*`Observable`实例在发送`OnCompleted`或`OnError`通知之前都会发出所有消息。如果有`OnError`通知，则会抛出异常（直接抛出`RuntimeException`异常，并将选中的异常包装在`RuntimeException`实例中）。

`toBlocking()`方法本身不阻塞，但返回的`BlockingObservable`实例的方法可能会阻塞。让我们看看其中的一些方法：

*   我们可以使用`forEach()`方法迭代`BlockingObservable`实例中的所有项。下面是使用此选项的示例：

    ```java
    Observable
      .interval(100L, TimeUnit.MILLISECONDS)
      .take(5)
      .toBlocking()
     .forEach(System.out::println);
    System.out.println("END");
    ```

这也是如何使*异步*代码同步*行为*的一个示例。由`interval()`方法创建的`Observable`实例将不在后台执行，因为`toBlocking()`方法使当前线程等待它完成。这就是为什么我们在这里使用`take(int)`方法，因为，否则*主*线程将永远被阻塞。`forEach()`方法将使用传递的函数打印五项，然后我们才能看到`END`输出。`BlockingObservable`类也有`toIterable()`方法。它返回的`Iterable`实例也可以用于迭代源发出的序列。

*   有类似于*异步*的*阻塞*方法，如`first()`、`last()`、`firstOrDefault()`和`lastOrDefault()`方法（我们在[第 4 章](4.html "Chapter 4. Transforming, Filtering, and Accumulating Your Data")、*转换、过滤和积累您的数据*中讨论过）。它们都在等待所需项目时阻塞。让我们来看看下面的代码片段：

这将打印`'3'`和`'15'`。

*   一个有趣的方法是`single()`方法；只有当*源*发出正好一项且*源完成*时，才返回一项。如果没有发出任何项，或者*源*发出多个项，则分别引发`NoSuchElementException`异常或`IllegalArgumentException`异常。
*   有一个`next()`方法不*阻塞*，而是返回一个`Iterable`实例。当从该`Iterable`实例中检索到`Iterator`实例时，其每个`next()`方法将*阻塞*，同时等待下一个传入项。这可以在无限`Observable`实例上使用，因为*当前线程*只有在等待*下一个*项时才会*阻塞*，然后才能继续。（请注意，如果在时间内没有人调用`next()`方法，则可能会跳过源元素）。下面是使用此选项的示例：

    ```java
    Iterable<Long> next = Observable
      .interval(100L, TimeUnit.MILLISECONDS)
      .toBlocking()
     .next();
    Iterator<Long> iterator = next.iterator();
    System.out.println(iterator.next());
    System.out.println(iterator.next());
    System.out.println(iterator.next());
    ```

*当前线程*将*阻塞*三次，持续 100 毫秒，`0`、`1`和`2`将在每次暂停后打印。有一个类似的方法叫做`latest()`，它返回一个`Iterable`实例。行为是不同的，因为`latest()`方法生成的`Iterable`实例返回源发出的最后一个项目，或者等待下一个项目（如果没有）。

```java
Iterable<Long> latest = Observable
  .interval(1000L, TimeUnit.MILLISECONDS)
  .toBlocking()
 .latest();
iterator = latest.iterator();
System.out.println(iterator.next());
Thread.sleep(5500L);
System.out.println(iterator.next());
System.out.println(iterator.next());
```

这将打印`0`，然后打印`5`和`6`。

### 注

可以在[查看/下载演示前面所有运算符以及聚合运算符的源代码 https://github.com/meddle0x53/learning-rxjava/blob/master/src/main/java/com/packtpub/reactive/chapter07/BlockingObservablesAndOperators.java](https://github.com/meddle0x53/learning-rxjava/blob/master/src/main/java/com/packtpub/reactive/chapter07/BlockingObservablesAndOperators.java) 。

使用`BlockingObservable`实例可以帮助我们收集测试数据。但是有一组名为**聚合运算符**的`Observable`运算符，当与`BlockingObservables`实例结合时，它们也很有用。

# 聚合运算符和 BlockingObservable 类

聚合操作符生成`Observable`实例，只发出一项并完成。此项由*源*`Observable`实例发出的所有项组成或计算。在本节中，我们将只讨论其中的两个。更多详细信息，请参考[https://github.com/ReactiveX/RxJava/wiki/Mathematical-and-Aggregate-Operators](https://github.com/ReactiveX/RxJava/wiki/Mathematical-and-Aggregate-Operators) 。

第一个操作符是`count()`或`countLong()`方法。发出*源*`Observable`实例发出的项数。例如：

```java
Observable
  .range(10, 100)
  .count()
  .subscribe(System.out::println);
```

这将打印`100`。

另一个是`toList()`或`toSortedList()`方法，它发出一个`list`变量（可以排序），包含*源*实例发出的所有项目，并完成。

```java
List<Integer> list = Observable
  .range(5, 15)
  .toList()
  .subscribe(System.out::println);
```

这将输出以下内容：

```java
[5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
```

所有这些方法，结合`toBlocking()`方法，都能很好地协同工作。例如，如果我们想要检索*源*`Observable`实例发出的所有项的列表，我们可以这样做：

```java
List<Integer> single = Observable
  .range(5, 15)
  .toList()
 .toBlocking().single();
```

我们可以使用我们想要的这个项目集合：例如用于测试。

### 提示

聚合操作符还包括一个`collect()`操作符，可以用于生成`Observable`实例和发出任意集合，例如`Set()`操作符。

# 使用聚合运算符和 BlockingObservable 类进行测试

使用前面两个部分中学习的操作符和方法，我们能够重新编写我们编写的测试，如下所示：

```java
@Test
public void testUsingBlockingObservable() {
  List<String> result = tested
    .toList()
 .toBlocking()
 .single();
  Assert.assertEquals(expected, result);
}
```

这里没有样板代码。我们检索作为列表发出的所有项，并将它们与预期列表进行比较。

在大多数情况下，使用`BlockingObsevables`类和聚合运算符非常有用。在测试发出长而慢序列的*异步*`Observable`实例时，它们并没有那么有用。长时间阻塞测试用例不是一个好的做法：缓慢的测试是糟糕的测试。

### 注

前面测试的源代码可以在[中找到 https://github.com/meddle0x53/learning-rxjava/blob/master/src/test/java/com/packtpub/reactive/chapter07/SortedObservableTest.java](https://github.com/meddle0x53/learning-rxjava/blob/master/src/test/java/com/packtpub/reactive/chapter07/SortedObservableTest.java) -这是第二种测试方法。

另一种测试方法没有帮助的情况是当我们想要检查*源*发送的`Notification`对象或订阅状态时。

还有一种编写测试的技术可以让我们对*订阅*本身进行更细粒度的控制，这是通过一种特殊的`Subscriber`-`TestSubscriber`实现的。

# 使用 TestSubscriber 类进行深入测试

`TestSubscriber`实例是一个特殊的`Subscriber`实例，可以传递给任何`Observable`实例的`subscribe()`方法。

我们可以从中检索所有收到的项目和通知。我们还可以查看最近收到通知的`thread`和订阅状态。

让我们使用它重新编写测试，以演示其功能和存储内容：

```java
@Test
public void testUsingTestSubscriber() {
  TestSubscriber<String> subscriber =
 new TestSubscriber<String>();
  tested.subscribe(subscriber);
  Assert.assertEquals(expected, subscriber.getOnNextEvents());
  Assert.assertSame(1, subscriber.getOnCompletedEvents().size());
  Assert.assertTrue(subscriber.getOnErrorEvents().isEmpty());
  Assert.assertTrue(subscriber.isUnsubscribed());
}
```

同样，测试非常简单。我们创建一个`TestSubscriber`实例，*用它订阅*到*测试的*`Observable`实例。`Observable`实例*完成*后，我们可以访问整个状态。让我们看看下面的术语列表：

*   通过`getOnNextEvents()`方法，我们能够检索`Observable`实例发出的所有项目，并将它们与*预期的*`List`变量进行比较。
*   通过`getOnCompletedEvents()`方法，我们可以检查*未完成的*通知，并检查其是否已发送。例如，`Observable.never()`方法不发送它。
*   通过`getOnErrorEvents()`方法，我们可以检查*或*错误通知是否存在。在本例中，我们*断言*不存在*错误*。
*   使用`isUnsubscribed()`方法，我们可以*断言*在*完成*之后，我们的`Subscriber`实例被*取消订阅*。

`TestSubscriber`实例也有一些*断言*方法。因此，还有一种编写测试的方法：

```java
@Test
public void testUsingTestSubscriberAssertions() {
  TestSubscriber<String> subscriber = new TestSubscriber<String>();
  tested.subscribe(subscriber);
 subscriber.assertReceivedOnNext(expected);
 subscriber.assertTerminalEvent();
 subscriber.assertNoErrors();
 subscriber.assertUnsubscribed();
}
```

这些几乎都是相同的*断言*，但都是通过`TestSubscriber`实例自己的`assert*`方法完成的。

### 注

前面测试的源代码可在[找到 https://github.com/meddle0x53/learning-rxjava/blob/master/src/test/java/com/packtpub/reactive/chapter07/SortedObservableTest.java](https://github.com/meddle0x53/learning-rxjava/blob/master/src/test/java/com/packtpub/reactive/chapter07/SortedObservableTest.java) -这是第三种和第四种测试方法。

通过这些技术，我们可以测试`RxJava`逻辑的不同行为和状态。在本章测试*异步*`Observable`实例中，还有最后一件事需要学习，例如通过`Observable.interval()`方法创建的实例。

# 借助 TestScheduler 类测试异步可观察实例

还有最后一种预定义的`scheduler`，我们在[第 6 章](6.html "Chapter 6. Using Concurrency and Parallelism with Schedulers")中没有提到，*使用与调度器*的并发性和并行性。这是`TestScheduler`调度器，一个`scheduler`设计用于单元测试。它上面调度的所有动作都被包装在包含它们应该在执行的时间的对象中，并且在调用`Scheduler`实例的`triggerActions()`方法之前不会执行。此方法执行所有未执行且计划在`Scheduler`实例当前时间或之前执行的操作。这次是虚拟的。这意味着它是由我们设定的，我们可以使用本`scheduler`的特殊方法前进到未来的任何时刻。

为了演示它，我们想开发另一种方法来创建一种新类型的`observable`。本章将不讨论方法本身的实现，但您可以在本书附带的源代码中找到它。

该方法创建一个`Observable`实例，以设置的时间间隔发送项目。但是间隔不是等距的，例如使用内置的`interval`方法。我们可以提供一个不同的多个*间隔*的列表，`Observable`实例将无限循环。该方法的签名如下：

```java
Observable<Long> interval(List<Long> gaps, TimeUnit unit, Scheduler scheduler)
```

如果我们传递一个只包含一个时间段值的`List`变量，则其行为应该与`Observable.interval`方法相同。下面是对这种情况的测试：

```java
@Test
public void testBehavesAsNormalIntervalWithOneGap() {
  TestScheduler testScheduler = Schedulers.test(); // (1)
  Observable<Long> interval = CreateObservable.interval(
 Arrays.asList(100L), TimeUnit.MILLISECONDS, testScheduler
 ); // (2)
  TestSubscriber<Long> subscriber = new TestSubscriber<Long>();
  interval.subscribe(subscriber); // (3)
  assertTrue(subscriber.getOnNextEvents().isEmpty()); // (4)
  testScheduler.advanceTimeBy(101L, TimeUnit.MILLISECONDS); // (5)
  assertEquals(Arrays.asList(0L), subscriber.getOnNextEvents());
  testScheduler.advanceTimeBy(101L, TimeUnit.MILLISECONDS); // (6)
  assertEquals(
    Arrays.asList(0L, 1L),
    subscriber.getOnNextEvents()
  );
  testScheduler.advanceTimeTo(1L, TimeUnit.SECONDS); // (7)
  assertEquals(
    Arrays.asList(0L, 1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L),
    subscriber.getOnNextEvents()
  );
}
```

让我们看一下下面的解释：

1.  我们使用`Schedulers.test()`方法创建`TestScheduler`实例。
2.  我们的方法接收一个`Scheduler`实例作为其第三个参数。它将*在其上发射项目*，因此我们传递`TestScheduler`实例。
3.  使用`TestSubscriber`实例，我们*订阅*到`Observable`实例。
4.  在订阅之后，我们不应该立即收到任何通知，所以我们会进行检查。
5.  `TestScheduler`实例有一个`advanceTimeBy(long, TimeUnit)`方法，它控制其`Worker`实例的时间，因此我们可以使用它来获得 101 毫秒的未来。101 毫秒后，我们希望收到一个项目-`0`。
6.  使用`advanceTimeBy()`方法，我们将时间提前 101 毫秒，我们应该已经收到`0`和`1`。
7.  `TestScheduler`实例的另一个重要方法是`advanceTimeTo(long, TimeUnit)`方法。它可以用于在将来提前到特定的时间点。因此，我们使用它来了解*订阅*的一秒钟已经过去的时刻。到那时，我们预计已经收到十份通知。

`TestScheduler`实例使用其`advanceTimeBy()`和`advanceTimeTo()`方法控制时间，因此我们不需要*阻塞*主实例`Thread`等待事件发生。我们可以回到已经发生的时间。对于`TestScheduler`实例，有一个全局事件顺序。因此，如果两个任务被安排在确切的同一时间，它们有一个执行顺序，并且可能会导致预期特定全局顺序的测试出现问题。如果我们有这样一个操作符要测试，我们应该通过计时到不同的值来避免这种情况，一个是 100 毫秒，另一个是 101 毫秒。使用这种技术，测试*异步*`Observable`实例不再是如此复杂的任务。

### 注

上述测试的源代码可在以下位置找到：[https://github.com/meddle0x53/learning-rxjava/blob/master/src/test/java/com/packtpub/reactive/chapter07/CreateObservableIntervalTest.java](https://github.com/meddle0x53/learning-rxjava/blob/master/src/test/java/com/packtpub/reactive/chapter07/CreateObservableIntervalTest.java) 。

# 总结

在本章中，我们不仅了解了如何使用 RxJava 编写程序，还了解了如何测试程序的各个方面。我们已经了解了一些新的操作符和`BlockingObservables`类。

RxJava 库中有许多本书中没有提到的操作符，但我们研究了更重要和有用的操作符。您可以随时参考[https://github.com/ReactiveX/RxJava/wiki](https://github.com/ReactiveX/RxJava/wiki) 其余部分。还有更多关于*订阅*、*背压*和`Observable`实例*生命周期*的内容，但根据您目前的知识，掌握库中的所有内容并不难。记住，这只是一个库，一个编写代码的工具。逻辑在这里是很重要的。这种编程方式与过程编程方式有些不同，但一旦你开始使用它，就会感觉很自然。

在下一章也是最后一章中，我们将学习如何释放由*订阅*分配的资源，如何防止内存泄漏，以及如何创建我们自己的操作符，这些操作符可以链接到`RxJava`逻辑中。